/********************************************************************************
 * Copyright (c) 2024 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_HEAVISIDE_H_
#define AIDGE_CORE_OPERATOR_HEAVISIDE_H_

#include <memory>
#include <string>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class HeavisideAttr {
    /**
     * @brief The value used in the output tensor when the input is 0.
     */
    Value
};

/**
 * @class Heaviside_Op
 * @brief Implements the Heaviside step function operation.
 *
 * The Heaviside step function computes element-wise output based on the input values:
 *
 * if x < 0: f(x) = 0
 * if x = 0: f(x) = Value
 * if x > 0: f(x) = 1
 *
 * The input and output Tensors have the same dimensions.
 *
 * @see OperatorTensor
 * @see Registrable
 */
class Heaviside_Op
    : public OperatorTensor,
      public Registrable<Heaviside_Op, std::string, std::function<std::shared_ptr<OperatorImpl>(const Heaviside_Op &)>> {

private:
    using Attributes_ = StaticAttributes<HeavisideAttr, float>;

    template <HeavisideAttr e>
    using attr = typename Attributes_::template attr<e>;

    const std::shared_ptr<Attributes_> mAttributes;

public:
    static const std::string Type;

    /**
     * @brief Constructor for the Heaviside operator.
     * @param[in] value The value to use in the output tensor when the input is 0.
     */
    Heaviside_Op(float value);

    /**
     * @brief Copy constructor.
     * @param op The Heaviside_Op to copy.
     * @details Copies the operator attributes and its output tensor(s), but not its input tensors.
     *          The new operator has no associated input.
     */
    Heaviside_Op(const Heaviside_Op &op);

    /**
     * @brief Clone the operator using its copy constructor.
     */
    std::shared_ptr<Operator> clone() const override;

    /**
     * @brief Set the backend for this operator.
     * @param name The backend name.
     * @param device The device index (default is 0).
     */
    void setBackend(const std::string &name, DeviceIdx_t device = 0) override;

    /**
     * @brief Get the set of available backends.
     * @return A set of backend names.
     */
    std::set<std::string> getAvailableBackends() const override;

    /**
     * @brief Get the input names required by this operator.
     * @return A vector containing the input names.
     */
    static const std::vector<std::string> getInputsName() {
        return {"data_input", "data_values"};
    }

    /**
     * @brief Get the output names generated by this operator.
     * @return A vector containing the output names.
     */
    static const std::vector<std::string> getOutputsName() {
        return {"output"};
    }

    /**
     * @brief Get the attributes of the operator.
     */
    inline std::shared_ptr<Attributes> attributes() const override {
        return mAttributes;
    }

    /**
     * @brief Get the value used for inputs equal to zero.
     * @return A reference to the value attribute.
     */
    inline float &value() const {
        return mAttributes->template getAttr<HeavisideAttr::Value>();
    }
};

/**
 * @brief Create a Heaviside node.
 *
 * Initializes a Heaviside node that computes the Heaviside step function for each element
 * of the input tensor, using the specified value for inputs equal to zero.
 *
 * @param value The value used in the output tensor when the input is 0.
 * @param name  Optional. The name of the node.
 *
 * @return A shared pointer to a Node representing the Heaviside operation.
 */
std::shared_ptr<Node> Heaviside(float value, const std::string &name = "");

} // namespace Aidge

namespace {
/**
 * @brief Define string representations for Heaviside attributes.
 */
template <>
const char *const EnumStrings<Aidge::HeavisideAttr>::data[] = {"value"};
}

#endif /* AIDGE_CORE_OPERATOR_HEAVISIDE_H_ */
