/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_OPERATOR_H_
#define AIDGE_CORE_OPERATOR_OPERATOR_H_

#include <cstddef>
#include <memory>
#include <string>
#include <vector>
#include <utility>

#ifdef PYBIND
#include <pybind11/pybind11.h>
#include <fmt/format.h>
#endif

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/utils/Attributes.hpp"
#include "aidge/utils/Types.h"

#ifdef PYBIND
namespace py = pybind11;
#endif
namespace Aidge {

/**
 * @enum OperatorType
 * @brief Specifies the type of the operator.
 */
enum class OperatorType {
    Data,   /**< Indicates a data operator. */
    Tensor  /**< Indicates a tensor operator. */
};

/**
 * @enum InputCategory
 * @brief Describes the category of an input for an operator.
 */
enum class InputCategory {
    Data,          /**< Regular data input. */
    Param,         /**< Parameter input. */
    OptionalData,  /**< Optional data input. */
    OptionalParam  /**< Optional parameter input. */
};

/**
 * @class Operator
 * @brief Base class for all operator types in the Aidge framework.
 *
 * The `Operator` class provides a foundation for implementing various operator types.
 * Derived classes must implement specific behaviors for computation, attributes,
 * and input/output handling.
 */
class Operator : public std::enable_shared_from_this<Operator> {
protected:
    /** @brief Implementation of the operator. */
    std::shared_ptr<OperatorImpl> mImpl;

    /** @brief Attributes of the associated Node.
     *
     * Default to empty vector for copy construction because two Operator cannot
     * be associated to the same Node.
     */
    std::vector<std::shared_ptr<DynamicAttributes>> mInheritedAttrs{};

private:
    /** @brief Type of the operator (e.g., "Add", "AveragePool"). */
    std::string mType;

    /** @brief Specifies whether the operator works on data or tensors. */
    const OperatorType mOperatorType;

    /** @brief Categories for each input. */
    const std::vector<InputCategory> mInputsCategory;

    /** @brief Number of outputs generated by the operator. */
    const IOIndex_t mNbOut;

    /** @brief Set of back edge input indexes for recurring operators. */
    std::set<IOIndex_t> mBackEdges;

public:
    /** @brief Deleted default constructor. */
    Operator() = delete;

    /**
     * @brief Constructs an Operator instance.
     *
     * @param[in] type           The type of operator (e.g., "Add", "AveragePool").
     * @param[in] inputsCategory Categories of each input.
     * @param[in] nbOut          Number of outputs.
     * @param[in] operatorType   Type of operator (Data or Tensor).
     */
    Operator(const std::string& type, const std::vector<InputCategory>& inputsCategory,
             const IOIndex_t nbOut, const OperatorType operatorType = OperatorType::Data)
             : mType(type),
            mOperatorType(operatorType),
            mInputsCategory(inputsCategory),
            mNbOut(nbOut)
    {
        // ctor
    }

    /**
     * @brief Copy constructor.
     *
     * @param[in] op The operator to copy.
     */
    Operator(const Operator& op):
        std::enable_shared_from_this<Operator>(),
        mOperatorType(op.mOperatorType),
        mInputsCategory(op.mInputsCategory),
        mNbOut(op.mNbOut),
        mBackEdges(op.mBackEdges)
    {
        mType = op.mType;
        mImpl = nullptr;
        // Implementation is never cloned. It is up to the non-abstract Operator copy-constructor to create a new implementation matching the copied Operator implementation.
        // See https://gitlab.eclipse.org/eclipse/aidge/aidge_core/-/merge_requests/8#note_1214050 for the discussion.
    }
    // std::shared_ptr<Operator> operator()(std::shared_ptr<DynamicAttributes> attrs) {
    //     mInheritedAttrs = attrs;
    //     return shared_from_this();
    // }

    /** @brief Virtual destructor. */
    virtual ~Operator() noexcept;

public:
    void setInheritedAttrs(std::shared_ptr<DynamicAttributes>& attr) {
        mInheritedAttrs.push_back(attr);
    }

    /**
     * @brief Creates a clone of the current operator.
     *
     * Derived classes must implement this method to provide a deep copy of the operator.
     *
     * @return A shared pointer to the cloned operator.
     */
    virtual std::shared_ptr<Operator> clone() const = 0;

    /** @brief Returns the attributes of the operator. */
    virtual std::shared_ptr<Attributes> attributes() const { return nullptr; };

    /**
     * @brief Get the currently associated Node's attributes.
     * @return Shared pointer to the Attributes of the associated Node.
     *
     * If no Node as be associated to the Operator, returns a `nullptr`.
     * @note As Operators have only been tested with a single associated Node,
     * only attributes of the first associated Node are returned. This should be
     * updated.
     */
    virtual std::shared_ptr<DynamicAttributes> inheritedAttributes() const {
        return mInheritedAttrs.empty() ? nullptr : mInheritedAttrs[0];
    }

    /**
     * @brief Associates a shallow copy of the specified input data with the operator.
     *
     * Derived classes must implement this method.
     *
     * @param[in] inputIdx Index of the input to associate.
     * @param[in] data     Data to associate.
     */
    virtual void associateInput(const IOIndex_t inputIdx, const std::shared_ptr<Data>& data) = 0;

    /** @brief Resets the specified input. */
    virtual void resetInput(const IOIndex_t inputIdx) = 0;

    /**
     * @brief Sets the specified input with a deep copy of the given data.
     *
     * Derived classes must implement this method.
     *
     * @param[in] inputIdx Index of the input to set.
     * @param[in] data     Data to set.
     */
    virtual void setInput(const IOIndex_t inputIdx, const std::shared_ptr<Data>& data) = 0;

    /** @brief Retrieves the raw input data for the specified index. */
    virtual std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const = 0;

    /**
     * @brief Sets the specified output with a deep copy of the given data.
     *
     * Derived classes must implement this method.
     *
     * @param[in] outputIdx Index of the output to set.
     * @param[in] data      Data to set.
     */
    virtual void setOutput(const IOIndex_t outputIdx, const std::shared_ptr<Data>& data) const = 0;

    /** @brief Retrieves the raw output data for the specified index. */
    virtual std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const = 0;

    /**
     * @brief Returns the backend implementation name.
     *
     * @return The name of the backend implementation.
     */
    virtual std::string backend() const noexcept { return mImpl ? mImpl->backend() : ""; }

    /**
     * @brief Sets the backend implementation.
     *
     * @param[in] name   Name of the backend.
     * @param[in] device Device index.
     */
    virtual void setBackend(const std::string& name, DeviceIdx_t device = 0) = 0;

    /**
     * @brief Sets the backend implementation for multiple devices.
     *
     * @param[in] backends A vector of backend and device index pairs.
     */
    void setBackend(const std::vector<std::pair<std::string, DeviceIdx_t>>& backends);

    virtual void setDataType(const DataType& dataType) const = 0;
    virtual void setDataFormat(const DataFormat& dataFormat) const = 0;
    /**
     * @brief Returns the available backend implementations.
     *
     * Derived classes must implement this method.
     *
     * @return A set of available backend names.
     */
    virtual std::set<std::string> getAvailableBackends() const = 0;
    /**
     * @brief Set a new OperatorImpl to the Operator
     *
     */
    inline void setImpl(std::shared_ptr<OperatorImpl> impl) { mImpl = impl; }

    /**
     * @brief Get the OperatorImpl of the Operator
     *
     */
    inline std::shared_ptr<OperatorImpl> getImpl() const noexcept {
        return mImpl;
    }

    /**
     * @brief Minimum amount of data from a specific input for one computation pass.
     * @param inputIdx Index of the input analyzed.
     * @return Elts_t
     */
    virtual Elts_t getNbRequiredData(const IOIndex_t inputIdx) const;

    // Amount of input data that cannot be overwritten during the execution.
    virtual Elts_t getNbRequiredProtected(const IOIndex_t inputIdx) const;

    // Memory required at an output for a given input size.
    virtual Elts_t getRequiredMemory(const IOIndex_t outputIdx, const std::vector<DimSize_t> &inputsSize) const;

    /**
     * @brief Total amount of consumed data from a specific input.
     *
     * @param inputIdx Index of the input analyzed.
     * @return Elts_t
     */
    virtual Elts_t getNbConsumedData(const IOIndex_t inputIdx) const;

    /**
     * @brief Total amount of produced data ready to be used on a specific output.
     *
     * @param outputIdx Index of the output analyzed.
     * @return Elts_t
     */
    virtual Elts_t getNbProducedData(const IOIndex_t outputIdx) const;

    virtual void updateConsummerProducer();

    virtual void resetConsummerProducer();

    virtual void forward();

    virtual void backward();
    /**
     * @brief Returns the type of the operator.
     *
     * @return The operator type as a string.
     */
    inline std::string type() const noexcept { return mType; }

    /**
     * @brief Returns the type of the operator (Data or Tensor).
     *
     * @return The operator type as an `OperatorType` enum value.
     */
    inline OperatorType operatorType() const noexcept { return mOperatorType; }

    /** @brief Returns the categories of all inputs. */
    inline std::vector<InputCategory> inputCategory() const { return mInputsCategory; }

    /** @brief Returns the category of a specific input. */
    inline InputCategory inputCategory(IOIndex_t idx) const {
        // AIDGE_ASSERT(idx < mInputsCategory.size(), "Input #{} out of range (number of inputs is {})", idx, mInputsCategory.size());
        return mInputsCategory.at(idx);
    }

    virtual inline bool isAtomic() const noexcept { return true; }

    /** @brief Returns the number of inputs. */
    inline IOIndex_t nbInputs() const noexcept { return mInputsCategory.size(); }

    /** @brief Returns the number of outputs. */
    inline IOIndex_t nbOutputs() const noexcept { return mNbOut; }

    /**
     * @brief Sets the back edge input indexes for recurring operators.
     *
     * @param[in] backEdges A set of input indexes representing back edges.
     */
    inline void setBackEdges(const std::set<IOIndex_t>& backEdges) { mBackEdges = backEdges; }

    /**
     * @brief Checks if a given input index is a back edge.
     *
     * @param[in] inputIdx Index of the input to check.
     * @return True if the input index is a back edge, false otherwise.
     */
    inline bool isBackEdge(IOIndex_t inputIdx) const { return mBackEdges.find(inputIdx) != mBackEdges.end(); }

    /** @brief Returns an empty vector of input names. */
    static const std::vector<std::string> getInputsName() { return {}; }

    /** @brief Returns an empty vector of output names. */
    static const std::vector<std::string> getOutputsName() { return {}; }

#ifdef PYBIND
    /**
     * @brief Returns a string representation of the operator (for Python bindings).
     */
    std::string repr() const {
        return fmt::format("Operator(type = '{}', nb_in = {}, nb_out = {}, attr = {}, backend = {})",
                    type(),
                    nbInputs(),
                    nbOutputs(),
                    (attributes() ? attributes()->repr() : "None"),
                    (mImpl ? "'"+backend()+"'" : "None"));
    }
#endif
};

}  // namespace Aidge

#endif  // AIDGE_CORE_OPERATOR_OPERATOR_H_
