/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/
#include <cstdint>  // std::int32_t
#include <memory>
#include <random>   // normal_distribution, uniform_real_distribution

#include "aidge/data/Tensor.hpp"
#include "aidge/filler/Filler.hpp"
#include "aidge/utils/Random.hpp"

template <typename T>
void Aidge::uniformFiller(std::shared_ptr<Aidge::Tensor> tensor, T min, T max) {
    AIDGE_ASSERT(tensor->getImpl(),
                 "Tensor got no implementation, cannot fill it.");
    AIDGE_ASSERT(NativeType_v<T> == tensor->dataType(), "Wrong data type {} and {}",NativeType_v<T>, tensor->dataType());


     using DistType = typename std::conditional<
        std::is_integral<T>::value,
        std::uniform_int_distribution<T>,
        std::uniform_real_distribution<T>
    >::type;

    DistType uniformDist(min, max);

    std::shared_ptr<Aidge::Tensor> cpyTensor;
    // Create cpy only if tensor not on CPU
    Aidge::Tensor& tensorWithValues =
        tensor->refCastFrom(cpyTensor, tensor->dataType(), "cpu");

    // Setting values
    for (std::size_t idx = 0; idx < tensorWithValues.size(); ++idx) {
        tensorWithValues.set<T>(idx, uniformDist(Aidge::Random::Generator::get()));
    }

    // Copy values back to the original tensors (actual copy only if needed)
    tensor->copyCastFrom(tensorWithValues);
}

template void Aidge::uniformFiller<float>(std::shared_ptr<Aidge::Tensor>, float,
                                          float);
template void Aidge::uniformFiller<double>(std::shared_ptr<Aidge::Tensor>,
                                           double, double);
template void Aidge::uniformFiller<std::int32_t>(std::shared_ptr<Aidge::Tensor>,
                                                 std::int32_t, std::int32_t);
template void Aidge::uniformFiller<std::int64_t>(std::shared_ptr<Aidge::Tensor>,
                                                 std::int64_t, std::int64_t);